
{TurboPascal 7.0 Programm zur Demonstration der COMx:-Port-Kontrolle}
{(c) alle Rechte bei Manfred Fischer, Heroldstr. 44, D-48163 Mnster}
{18.Mrz 1995}

Program TestCom;

Uses Crt, ScreenIO;


{die Unit RS232 ist hier komplett eingebunden, um CTS, DCD etc. direkt}
{auswerten zu knnen.}
CONST  {Offsets zum Inhalt der Variablen gPort0:}
    LineControlRegister             = 3; {d.h. bei COM1: $3FB}
    ProgrammableBaudGenerator_LSB   = 0; {               $3F8}
    ProgrammableBaudGenerator_MSB   = 1; {               $3F9}
    LineStatusRegister              = 5; {               $3FD}
    InterruptIdentificationRegister = 2; {               $3FA}
    InterruptEnableRegister         = 1; {               $3F9}
    ModemControlRegister            = 4; {               $3FC}
    ModemStatusRegister             = 6; {               $3FE}
    ScratchPadRegister              = 7; {               $3FF}
    ReceiveBufferRegister           = 0; {               $3F8}
    TransmitterHoldingRegister      = 0; {               $3F8}

Type SystemString = String[25];

Var
  gPort0 : Word; {globale Variable, enthlt Adresse des ersten Ports}


{--- Bit Manipulationsfunktionen ---}
{Bits gezhlt von rechts von 0 bis 15}

Function GetBit(b : Word; Nr : Byte) : Byte;
begin {Bits gezhlt von 0 bis 15}
  GetBit := Ord(Odd(b SHR Nr)); {liefert 0 oder 1}
end;

Function IsSetBit(b : Word; Nr : Byte) : Boolean;
begin {analog zu GetBit()}
  IsSetBit := Odd(b SHR Nr); {d.h. Odd(0) = FALSE}
end;

Function SetBit(b : Word; Nr : Byte) : Word; {Bit setzen}
begin
  SetBit := b or (1 SHL Nr);
end;

Function BlankBit(b : Word; Nr : Byte) : Word; {Bit lschen}
begin
  BlankBit := b and not (1 SHL Nr);
end;



{--- Anzahl der COMx:-Schnittstellen und ihre Port-Adressen ---}


Const
  BiosRamSeg = $0040; {Bereich des PC-RAM's, in dem das BIOS Werte speichert}

Function Equipment : Word; {Zustand des Hardware-Equipment}
begin
  Equipment := MemW[BiosRamSeg:$0010];
end;

Function ComPorts : Byte; {Anzahl der COMx:-Schnittstellen}
begin
  ComPorts := (Equipment SHR 9) AND 3;
end;

Function ComPortAddr(ComNo : Byte) : Word;
begin {liefert den Wert 0, wenn nicht vorhanden}
  if (ComNo >= 1) and (ComNo <= ComPorts) then
    ComPortAddr := MemW[BiosRamSeg:$0000 + (Pred(ComNo) * 2)]
  else ComPortAddr := 0; {d.h. nicht vorhanden}
end;

Function DirectComPortAddr(ComNo : Byte) : Word;
begin {ComNo im Bereich 1..4, ohne Kontrolle auf Plausibilitt}
  DirectComPortAddr := MemW[BiosRamSeg:$0000 + (Pred(ComNo) * 2)]
end;



{--- Nun die allgemeinen Routinen mit den Port-Zugriffen ---}
{--- bentigen die globale Variable gPort0 ---}

{--- Line Control Register ---}

Function GetDatenBits : Byte;
Var i : Integer;
begin
  i := Port[gPort0 + LineControlRegister];
  GetDatenBits:=GetBit(i,1)*2 + GetBit(i,0) + 5; {DatenBits mglich von 5 bis 8}
end;

Procedure SetDatenBits(Anzahl : Byte);
Var i : integer;
begin
  i := Port[gPort0 + LineControlRegister];
  Anzahl := (Anzahl - 5) MOD 4;    {nur 2 Bit breit !}
  i := i and (not 3);              {Bit 0 und 1 lschen}
  i := i + Anzahl;                 {Bit 0 und 1 entsprechend neu setzen}
  Port[gPort0 + LineControlRegister] := i;
end;

Function GetStopBits : Byte;
begin {mgliche Anzahl: 1 oder 2 StopBits}
  GetStopBits := GetBit(Port[gPort0 + LineControlRegister],2) + 1;
end;

Procedure SetStopBits(Anzahl : Byte);
begin {mgliche Anzahl: 1 oder 2 StopBits}
  if ((Anzahl MOD 2) = 0) then
    Port[gPort0 + LineControlRegister] :=
         SetBit(Port[gPort0 + LineControlRegister],2)
  else Port[gPort0 + LineControlRegister] :=
         BlankBit(Port[gPort0 + LineControlRegister],2);
end;

Function GetParity : Char;
begin
  if (GetBit(Port[gPort0 + LineControlRegister],3) = 1) then
  begin   {ParityEnableBit ist gesetzt}
    case GetBit(Port[gPort0 + LineControlRegister],4) of
      0 : GetParity := 'O'; (* Odd *)
      1 : GetParity := 'E'; (* Even *)
    end;
  end
  else GetParity := 'N'; (* no parity *)
end;

Procedure SetParity(P : Char);
Var i : Integer;
begin
  i := Port[gPort0 + LineControlRegister];
  if (UpCase(P) = 'N') then Port[gPort0 + LineControlRegister] := BlankBit(i,3)
  else
  begin
    i := SetBit(i,3);
    case UpCase(P) of
      'O': i := BlankBit(i,4);
      'E': i := SetBit(i,4);
    end;
    Port[gPort0 + LineControlRegister] := i;
  end;
end;

(* stick parity bit and break control bit not managed here *)

Procedure SwitchToBaudGenerator;
{Set the Divisor Latch Access Bit (DLAB) for $3F8 and $3F9}
begin
  Port[gPort0 + LineControlRegister] :=
    SetBit(Port[gPort0 + LineControlRegister],7);
end;

Procedure TurnOffBaudGenerator;
{Turn OFF the Divisor Latch Access Bit (DLAB) for $3F8 and $3F9}
begin
  Port[gPort0 + LineControlRegister] :=
    BlankBit(Port[gPort0 + LineControlRegister],7);
end;


{--- Programmable Baud Generator ---}
{gltige Baudraten beim 8250: 50,110,134.5,150,300,600,1200,1800}
{2000,2400,3600,4800,7200,9600. Bei neueren Prozessoren auch hher}

Function GetBaudRate : Word;
Var
  w : Word;
  r : Real;
begin   {DLAB = 1 !}
  w := (Port[gPort0 + ProgrammableBaudGenerator_MSB] shl 8) +
        Port[gPort0 + ProgrammableBaudGenerator_LSB];
  r := (9600.0 / w) * 12.0;
  GetBaudRate := Round(r);
end;

Procedure SetBaudRate(Baud : Word);
Var Divisor : Real;
    d : Word;
begin     {DLAB = 1 !}
  case    Baud of
    50..9600 : ; {da gltig}
    else Baud:=300;
  end;
  Divisor := (9600.0 / Baud) * 12.0;
  d := Round(Divisor);
  case d of
     56..58  : d := 58;  {2000 Baud}
    800..900 : d := 857; {134.5 Baud}
    else if (d < 12) then d := 12; {nie grer als 9600 Baud}
  end;
  Port[gPort0 + ProgrammableBaudGenerator_LSB] := d MOD 256;
  Port[gPort0 + ProgrammableBaudGenerator_MSB] := d shr 8;
end;


{--- Line Status Register    (read only !) ---}


Function NewByteIsThere : Boolean;
begin
  NewByteIsThere := IsSetBit(Port[gPort0 + LineStatusRegister],0);
end;

Function AnyError : Boolean;
begin
  {bit 1: OverRun Error, empfangenes Zeichen wurde berschrieben}
  {bit 2: Parity Error}
  {bit 3: Framing Error, Stopbits falsch}
  AnyError := ((Port[gPort0 + LineStatusRegister] SHR 1) MOD 8) <> 0;
end;

Function ReadyForNewByte : Boolean;
begin
  ReadyForNewByte := IsSetBit(Port[gPort0 + LineStatusRegister],6);
end;


{--- Receiving and Transmitting ---}

Procedure TransmittChar(c : Char);
begin        {DLAB = 0 !}
  Port[gPort0 + TransmitterHoldingRegister] := Ord(c);
end;

Function ReceiveByte : Byte;
begin        {DLAB = 0 !}
  ReceiveByte := Port[gPort0 + ReceiveBufferRegister];
end;

Function ReceiveChar : Char;
begin        {DLAB = 0 !}
  ReceiveChar := Chr(Port[gPort0 + ReceiveBufferRegister]);
end;


{--- Modem Status Register ---}

Function ClearToSend : Boolean;
begin      {CTS = Verbindung besteht}
  ClearToSend := IsSetBit(Port[gPort0 + ModemStatusRegister],4);
end;

Function DataSetReady : Boolean;
begin      {DSR = Modem hat genug Saft und ist bereit}
  DataSetReady := IsSetBit(Port[gPort0 + ModemStatusRegister],5);
end;

Function RingIndicator : Boolean;
begin      {RI, hier nicht benutzt, nur bei Pufferung in Gebrauch}
  RingIndicator := IsSetBit(Port[gPort0 + ModemStatusRegister],6);
end;

Function DataCarrierDetect : Boolean;
begin {DCD = Empfangssignalpegel, Lautstrke ausreichend}
  DataCarrierDetect := IsSetBit(Port[gPort0 + ModemStatusRegister],7);
end;


{--- Modem Control Register ---}

Procedure SetDataTerminalReady;
begin
  Port[gPort0 + ModemControlRegister] :=
       SetBit(Port[gPort0 + ModemControlRegister],0);
end;

Procedure BlankDataTerminalReady;
begin
  Port[gPort0 + ModemControlRegister] :=
       BlankBit(Port[gPort0 + ModemControlRegister],0);
end;

Procedure SetRequestToSend;
begin  {RTS = Anfrage, ob Sendebereitschaft vorliegt}
  Port[gPort0 + ModemControlRegister] :=
       SetBit(Port[gPort0 + ModemControlRegister],1);
end;

Procedure BlankRequestToSend;
begin
  Port[gPort0 + ModemControlRegister] :=
       BlankBit(Port[gPort0 + ModemControlRegister],1);
end;

Function DataTerminalReady : Boolean;
begin
  DataTerminalReady :=
    IsSetBit(Port[gPort0 + ModemControlRegister],0);
end;

Function RequestToSend : Boolean;
begin
  RequestToSend :=
    IsSetBit(Port[gPort0 + ModemControlRegister],1);
end;


{--- Zusammenfassung ---}

Procedure InitOfRS232(Baud : Word; DatenBit : Byte;
                      Parity : Char; StopBit : Byte);
begin
  SwitchToBaudGenerator;
  SetBaudRate(Baud);
  SetDatenBits(DatenBit);
  SetParity(Parity);
  SetStopBits(StopBit);
  TurnOffBaudGenerator;
end;

Procedure ReadStatusOfRS232(var Baud : Word; var DatenBit : Byte;
                            var Parity : char; var StopBit : Byte);
begin
  SwitchToBaudGenerator;
  Baud     := GetBaudRate;
  DatenBit := GetDatenBits;
  Parity   := GetParity;
  StopBit  := GetStopBits;
  TurnOffBaudGenerator;
end;

Procedure ModemAn;
begin
  SetDataTerminalReady;
  SetRequestToSend;
end;

Procedure ModemAus;
begin
  BlankDataTerminalReady;
  BlankRequestToSend;
end;



{--- Nun die Routinen zur Ansprache mit eigenem System ---}


Procedure InitCom(ComNo : Byte;        {ComNo 1..4}
                  Baud : Word; Data : Byte;
                  Parity : Char; Stop : Byte);
begin
  gPort0 := DirectComPortAddr(ComNo); {COM1=$40:00, COM2=$40:02 ..}
  InitOfRS232(Baud,Data,Parity,Stop);
  ModemAn;
end;

Procedure SendChar(ComNo : Byte; C : Char);
begin
  gPort0 := DirectComPortAddr(ComNo);
  TransmittChar(C);
end;

Function GetChar(ComNo : Byte) : Char;
begin
  gPort0 := DirectComPortAddr(ComNo);
  GetChar := ReceiveChar;
end;

Function CharThere(ComNo : Byte) : Boolean;
begin
  gPort0 := DirectComPortAddr(ComNo);
  CharThere := NewByteIsThere;
end;

Function ComReady(ComNo : Byte) : Boolean;
begin
  gPort0 := DirectComPortAddr(ComNo);
  ComReady := ReadyForNewByte;
end;

Function IdentifyComStatus(ComNo : Byte) : SystemString;
Var Erg, Temp : SystemString;
    Baud : Word;
    DatenBit, StopBit : Byte;
    Par : Char;
begin
  gPort0 := DirectComPortAddr(ComNo);
  ReadStatusOfRS232(Baud,DatenBit,Par,StopBit);
  Str(Baud,Erg);
  Str(DatenBit,Temp);
  Erg := Erg + ',' + Temp + ',' + Par;
  Str(StopBit,Temp);
  Erg := Erg + ',' + Temp + '  ';
  if ClearToSend then Erg := Erg + ' CTS';
  if DataSetReady then Erg := Erg + ' DSR';
  if DataCarrierDetect then Erg := Erg + ' CD';
  IdentifyComStatus := Erg;
end;

{-----------------------------------------------------}

Procedure ModemStatusAnzeigen(ComPort : Byte; {welche Schnittstelle}
                              X0, Y0 : Word); {Fenster Koordinaten links oben}
Const StdStift  = 14;
      StdPapier = 3;
      Schwarz   = 0;
      Breite = 38;
      Hoehe  = 9;

Type String3  = String[3];
     String35 = String[35];

Var OldWindMin, OldWindMax : Word;
    Ch : Char;
    CTS, DSR, DCD, DTR, RTS : Boolean;

  Procedure Leitung(Abkz : String3; Txt : String35; Status : Boolean);
  begin
    if Status then ScreenIO.Farben(StdStift,StdPapier)
    else ScreenIO.Farben(Schwarz,StdPapier);
    write(Abkz:4);
    if Status then ScreenIO.Farben(Schwarz,StdPapier)
    else ScreenIO.Farben(StdStift,StdPapier);
    writeln('  ',Txt)
  end;

begin
  OldWindMin := WindMin;
  OldWindMax := WindMax;
  ScreenIO.MakeFenster(X0,Y0,X0+Breite-1,Y0+Hoehe-1,2);
  GotoXY(X0+1,Y0);
  write('Status des Modems an COM'+Chr(ComPort+48)+':');
  GotoXY(X0+1,Y0+Hoehe-1);
  write('Ende mit <ESC>');
  Window(X0+1,Y0+1,X0+Breite-2,Y0+Hoehe-2);

  gPort0 := DirectComPortAddr(ComPort);

  repeat
    ClrScr;
    CTS := ClearToSend;
    DSR := DataSetReady;
    DCD := DataCarrierDetect;
    DTR := DataTerminalReady;
    RTS := RequestToSend;
    if DTR then Leitung('DTR','Modem angefordert',DTR)
    else Leitung('---','Modem nicht angefordert',DTR);
    if RTS then Leitung('RTS','Leitung angefordert',RTS)
    else Leitung('---','Leitung nicht angefordert',RTS);
    if CTS then Leitung('CTS','Verbindung besteht',CTS)
    else Leitung('---','Verbindung ist unterbrochen',CTS);
    if DSR then Leitung('DSR','Modem ist eingeschaltet',DSR)
    else Leitung('---','Modem ist ausgeschaltet',DSR);
    if DCD then Leitung('CD ','Lautstrke ist ausreichend',DCD)
    else Leitung('-- ','zu leises Empfangssignal',DCD);
    ScreenIO.Farben(Schwarz,StdPapier);
    writeln;
    write('  nochmalige Abfrage mit <CR>');
    ScreenIO.ReadUpCaseKey(Ch);
  until (Ch = #27);

  WindMin := OldWindMin;
  WindMax := OldWindMax;
end;

Var S : String;
begin {Testen von COMx:,  x = ParamStr(1)}
  if (ParamCount = 0)
  then writeln('Aufruf mit Parameter: C>testcom <Nummer des ComPorts>')
  else
  begin
    S := ParamStr(1);
    if not (S[1] in ['1'..'8']) then S[1] := '1';
    ModemStatusAnzeigen(Ord(S[1]) - Ord('0'),25,5);
  end;
end.

