

{Datei: RS232.Pas}
{TurboPascal 4.0 - 7.0 Unit -- Alle Rechte bei Manfred Fischer, D-48163 Mnster}
{29.01.1995 *** entwickelt aus Komponenten der Units LowMem4 und Aux8250}
{Ansteuerung der RS-232-Schnittstellen COM1: bis COM4: an 80x86-PC's zur}
{seriellen, ungepufferten Kommunikation mit dem Prozessor 8250 bei direktem}
{Port-Zugriff. Der 8250 ist seit den PC-Urzeiten implementiert, neue PC's}
{enthalten andere Kommunikations-Prozessoren, die zum 8250 abwrtskompatibel}
{sind.}
{letzte nderung: 29.06.1995: Debugging, nun mit CTS-Abfrage}

Unit RS232; {$A+,S-,I-,R-,B-,D-,E-,F-,L-,N-,O-,V-}

Interface

Type
  SystemString = String[25];

Function ComPorts : Byte; {Anzahl der COMx:-Schnittstellen}
Function ComPortAddr(ComNo : Byte) : Word;

Procedure InitCom(ComNo : Byte;        {ComNo 1..4}
                  Baud : Word; Data : Byte; Parity : Char; Stop : Byte);
Procedure SendChar(ComNo : Byte; C : Char);
Function  GetChar(ComNo : Byte) : Char;
Function  CharThere(ComNo : Byte) : Boolean;
Function  ComReady(ComNo : Byte) : Boolean;
Function  IdentifyComStatus(ComNo : Byte) : SystemString;


Implementation

{--------------------------------------------------------------
Diese Routinen beziehen sich auf den seriellen Kommunikations-
prozessor 8250 der RS232/V.24 Schnittstelle unter der CPU 80x86.

Der 8250 ist default-maessig in allen PC's eingebaut.
Der 8250 verfuegt ueber mehrere Register, einige ($3F8,$3F9 bei COM1:,
$2F8,$2F9 bei COM2: etc.) haben mehrere Funktionen:
das Divisor Latch Access Bit (DLAB) im Line Control Register
steuert die Benutzung der Regs $3F8,$3F9 (bei COM1:). Voreingestellt
ist die Benutzung der beiden Regs als programmierbarer
Baud-Generator. Deshalb muss in jedem Fall vor der eigentlichen
Uebertragung das DLAB geloescht werden (procedure TurnOffBaudGenerator).
Die Benutzung von Reg $3F8 erfolgt dann folgendermassen:
wird hineingeschrieben, dient es als Senderegister,
kommt von "aussen" ein Zeichen, dient es als Empfangsregister.
Die Bereitschaft zu der jeweiligen Arbeitsweise regelt das
Line Status Register. Der Leitungszustand und die Bereitschaft
des Modems ist ueber das Modem Status Register abfragbar.

Die Adressen der jeweils ersten Adapter-Register an der CPU (= Port-Adressen)
finden sich im BIOS-RAM ab $0040:0000. Portadressen $3F8-$3FF beziehen sich
auf COM1:, $2F8-$2FF beziehen sich auf COM2:, etc., falls vorhanden.
}

CONST  {Offsets zum Inhalt der Variablen gPort0:}
    LineControlRegister             = 3; {d.h. bei COM1: $3FB}
    ProgrammableBaudGenerator_LSB   = 0; {               $3F8}
    ProgrammableBaudGenerator_MSB   = 1; {               $3F9}
    LineStatusRegister              = 5; {               $3FD}
    InterruptIdentificationRegister = 2; {               $3FA}
    InterruptEnableRegister         = 1; {               $3F9}
    ModemControlRegister            = 4; {               $3FC}
    ModemStatusRegister             = 6; {               $3FE}
    ScratchPadRegister              = 7; {               $3FF}
    ReceiveBufferRegister           = 0; {               $3F8}
    TransmitterHoldingRegister      = 0; {               $3F8}

Var
  gPort0 : Word; {globale Variable, enthlt Adresse des ersten Ports}



{--- Bit Manipulationsfunktionen ---}
{Bits gezhlt von rechts von 0 bis 15}

Function GetBit(b : Word; Nr : Byte) : Byte;
begin {Bits gezhlt von 0 bis 15}
  GetBit := Ord(Odd(b SHR Nr)); {liefert 0 oder 1}
end;

Function IsSetBit(b : Word; Nr : Byte) : Boolean;
begin {analog zu GetBit()}
  IsSetBit := Odd(b SHR Nr); {d.h. Odd(0) = FALSE}
end;

Function SetBit(b : Word; Nr : Byte) : Word; {Bit setzen}
begin
  SetBit := b or (1 SHL Nr);
end;

Function BlankBit(b : Word; Nr : Byte) : Word; {Bit lschen}
begin
  BlankBit := b and not (1 SHL Nr);
end;



{--- Anzahl der COMx:-Schnittstellen und ihre Port-Adressen ---}


Const
  BiosRamSeg = $0040; {Bereich des PC-RAM's, in dem das BIOS Werte speichert}

Function Equipment : Word; {Zustand des Hardware-Equipment}
begin
  Equipment := MemW[BiosRamSeg:$0010];
end;

Function ComPorts : Byte; {Anzahl der COMx:-Schnittstellen}
begin
  ComPorts := (Equipment SHR 9) AND 3;
end;

Function ComPortAddr(ComNo : Byte) : Word;
begin {liefert den Wert 0, wenn nicht vorhanden}
  if (ComNo >= 1) and (ComNo <= ComPorts) then
    ComPortAddr := MemW[BiosRamSeg:$0000 + (Pred(ComNo) * 2)]
  else ComPortAddr := 0; {d.h. nicht vorhanden}
end;

Function DirectComPortAddr(ComNo : Byte) : Word;
begin {ComNo im Bereich 1..4, ohne Kontrolle auf Plausibilitt}
  DirectComPortAddr := MemW[BiosRamSeg:$0000 + (Pred(ComNo) * 2)]
end;



{--- Nun die allgemeinen Routinen mit den Port-Zugriffen ---}
{--- bentigen die globale Variable gPort0 ---}

{--- Line Control Register ---}

Function GetDatenBits : Byte;
Var i : Integer;
begin
  i := Port[gPort0 + LineControlRegister];
  GetDatenBits:=GetBit(i,1)*2 + GetBit(i,0) + 5; {DatenBits mglich von 5 bis 8}
end;

Procedure SetDatenBits(Anzahl : Byte);
Var i : integer;
begin
  i := Port[gPort0 + LineControlRegister];
  Anzahl := (Anzahl - 5) MOD 4;    {nur 2 Bit breit !}
  i := i and (not 3);              {Bit 0 und 1 lschen}
  i := i + Anzahl;                 {Bit 0 und 1 entsprechend neu setzen}
  Port[gPort0 + LineControlRegister] := i;
end;

Function GetStopBits : Byte;
begin {mgliche Anzahl: 1 oder 2 StopBits}
  GetStopBits := GetBit(Port[gPort0 + LineControlRegister],2) + 1;
end;

Procedure SetStopBits(Anzahl : Byte);
begin {mgliche Anzahl: 1 oder 2 StopBits}
  if ((Anzahl MOD 2) = 0) then
    Port[gPort0 + LineControlRegister] :=
         SetBit(Port[gPort0 + LineControlRegister],2)
  else Port[gPort0 + LineControlRegister] :=
         BlankBit(Port[gPort0 + LineControlRegister],2);
end;

Function GetParity : Char;
begin
  if (GetBit(Port[gPort0 + LineControlRegister],3) = 1) then
  begin   {ParityEnableBit ist gesetzt}
    case GetBit(Port[gPort0 + LineControlRegister],4) of
      0 : GetParity := 'O'; (* Odd *)
      1 : GetParity := 'E'; (* Even *)
    end;
  end
  else GetParity := 'N'; (* no parity *)
end;

Procedure SetParity(P : Char);
Var i : Integer;
begin
  i := Port[gPort0 + LineControlRegister];
  if (UpCase(P) = 'N') then Port[gPort0 + LineControlRegister] := BlankBit(i,3)
  else
  begin
    i := SetBit(i,3);
    case UpCase(P) of
      'O': i := BlankBit(i,4);
      'E': i := SetBit(i,4);
    end;
    Port[gPort0 + LineControlRegister] := i;
  end;
end;

(* stick parity bit and break control bit not managed here *)

Procedure SwitchToBaudGenerator;
{Set the Divisor Latch Access Bit (DLAB) for $3F8 and $3F9}
begin
  Port[gPort0 + LineControlRegister] :=
    SetBit(Port[gPort0 + LineControlRegister],7);
end;

Procedure TurnOffBaudGenerator;
{Turn OFF the Divisor Latch Access Bit (DLAB) for $3F8 and $3F9}
begin
  Port[gPort0 + LineControlRegister] :=
    BlankBit(Port[gPort0 + LineControlRegister],7);
end;


{--- Programmable Baud Generator ---}
{gltige Baudraten beim 8250: 50,110,134.5,150,300,600,1200,1800}
{2000,2400,3600,4800,7200,9600. Bei neueren Prozessoren auch hher}

Function GetBaudRate : Word;
Var
  w : Word;
  r : Real;
begin   {DLAB = 1 !}
  w := (Port[gPort0 + ProgrammableBaudGenerator_MSB] shl 8) +
        Port[gPort0 + ProgrammableBaudGenerator_LSB];
  r := (9600.0 / w) * 12.0;
  GetBaudRate := Round(r);
end;

Procedure SetBaudRate(Baud : Word);
Var Divisor : Real;
    d : Word;
begin     {DLAB = 1 !}
  case    Baud of
    50..9600 : ; {da gltig}
    else Baud:=300;
  end;
  Divisor := (9600.0 / Baud) * 12.0;
  d := Round(Divisor);
  case d of
     56..58  : d := 58;  {2000 Baud}
    800..900 : d := 857; {134.5 Baud}
    else if (d < 12) then d := 12; {nie grer als 9600 Baud}
  end;
  Port[gPort0 + ProgrammableBaudGenerator_LSB] := d MOD 256;
  Port[gPort0 + ProgrammableBaudGenerator_MSB] := d shr 8;
end;


{--- Line Status Register    (read only !) ---}


Function NewByteIsThere : Boolean;
begin
  NewByteIsThere := IsSetBit(Port[gPort0 + LineStatusRegister],0);
end;

Function AnyError : Boolean;
begin
  {bit 1: OverRun Error, empfangenes Zeichen wurde berschrieben}
  {bit 2: Parity Error}
  {bit 3: Framing Error, Stopbits falsch}
  AnyError := ((Port[gPort0 + LineStatusRegister] SHR 1) MOD 8) <> 0;
end;

Function ReadyForNewByte : Boolean;
begin
  ReadyForNewByte := (Port[gPort0 + LineStatusRegister] AND 96) = 96;
  {d.h. Transmitterholdingregister ist bereit und kein Fehler!}
end;


{--- Receiving and Transmitting ---}

Procedure TransmittChar(c : Char);
begin        {DLAB = 0 !}
  Port[gPort0 + TransmitterHoldingRegister] := Ord(c);
end;

Function ReceiveByte : Byte;
begin        {DLAB = 0 !}
  ReceiveByte := Port[gPort0 + ReceiveBufferRegister];
end;

Function ReceiveChar : Char;
begin        {DLAB = 0 !}
  ReceiveChar := Chr(Port[gPort0 + ReceiveBufferRegister]);
end;


{--- Modem Status Register ---}

Function ClearToSend : Boolean;
begin      {CTS = Verbindung besteht}
  ClearToSend := IsSetBit(Port[gPort0 + ModemStatusRegister],4);
end;

Function DataSetReady : Boolean;
begin      {DSR = Modem hat genug Saft und ist bereit}
  DataSetReady := IsSetBit(Port[gPort0 + ModemStatusRegister],5);
end;

Function RingIndicator : Boolean;
begin      {RI, hier nicht benutzt, nur bei Pufferung in Gebrauch}
  RingIndicator := IsSetBit(Port[gPort0 + ModemStatusRegister],6);
end;

Function DataCarrierDetect : Boolean;
begin {DCD = Empfangssignalpegel, Lautstrke ausreichend}
  DataCarrierDetect := IsSetBit(Port[gPort0 + ModemStatusRegister],7);
end;


{--- Modem Control Register ---}

Procedure SetDataTerminalReady;
begin
  Port[gPort0 + ModemControlRegister] :=
       SetBit(Port[gPort0 + ModemControlRegister],0);
end;

Procedure BlankDataTerminalReady;
begin
  Port[gPort0 + ModemControlRegister] :=
       BlankBit(Port[gPort0 + ModemControlRegister],0);
end;

Procedure SetRequestToSend;
begin  {RTS = Anfrage, ob Sendebereitschaft vorliegt}
  Port[gPort0 + ModemControlRegister] :=
       SetBit(Port[gPort0 + ModemControlRegister],1);
end;

Procedure BlankRequestToSend;
begin
  Port[gPort0 + ModemControlRegister] :=
       BlankBit(Port[gPort0 + ModemControlRegister],1);
end;

Function DataTerminalReady : Boolean;
begin
  DataTerminalReady :=
    IsSetBit(Port[gPort0 + ModemControlRegister],0);
end;

Function RequestToSend : Boolean;
begin
  RequestToSend :=
    IsSetBit(Port[gPort0 + ModemControlRegister],1);
end;


{--- Zusammenfassung ---}

Procedure InitOfRS232(Baud : Word; DatenBit : Byte;
                      Parity : Char; StopBit : Byte);
begin
  SwitchToBaudGenerator;
  SetBaudRate(Baud);
  SetDatenBits(DatenBit);
  SetParity(Parity);
  SetStopBits(StopBit);
  TurnOffBaudGenerator;
end;

Procedure ReadStatusOfRS232(var Baud : Word; var DatenBit : Byte;
                            var Parity : char; var StopBit : Byte);
begin
  SwitchToBaudGenerator;
  Baud     := GetBaudRate;
  DatenBit := GetDatenBits;
  Parity   := GetParity;
  StopBit  := GetStopBits;
  TurnOffBaudGenerator;
end;

Procedure ModemAn;
begin
  SetDataTerminalReady;
  SetRequestToSend;
end;

Procedure ModemAus;
begin
  BlankDataTerminalReady;
  BlankRequestToSend;
end;



{--- Nun die Routinen zur Ansprache mit eigenem System ---}


Procedure InitCom(ComNo : Byte;        {ComNo 1..4}
                  Baud : Word; Data : Byte;
                  Parity : Char; Stop : Byte);
begin
  gPort0 := DirectComPortAddr(ComNo); {COM1=$40:00, COM2=$40:02 ..}
  InitOfRS232(Baud,Data,Parity,Stop);
  ModemAn;
end;

Procedure SendChar(ComNo : Byte; C : Char);
begin
  gPort0 := DirectComPortAddr(ComNo);
  TransmittChar(C);
end;

Function GetChar(ComNo : Byte) : Char;
begin
  gPort0 := DirectComPortAddr(ComNo);
  GetChar := ReceiveChar;
end;

Function CharThere(ComNo : Byte) : Boolean;
begin
  gPort0 := DirectComPortAddr(ComNo);
  CharThere := NewByteIsThere;
end;

Function ComReady(ComNo : Byte) : Boolean;
begin
  gPort0 := DirectComPortAddr(ComNo);
  ComReady := ReadyForNewByte and ClearToSend;
end;

Function IdentifyComStatus(ComNo : Byte) : SystemString;
Var Erg, Temp : SystemString;
    Baud : Word;
    DatenBit, StopBit : Byte;
    Par : Char;
begin
  gPort0 := DirectComPortAddr(ComNo);
  ReadStatusOfRS232(Baud,DatenBit,Par,StopBit);
  Str(Baud,Erg);
  Str(DatenBit,Temp);
  Erg := Erg + ',' + Temp + ',' + Par;
  Str(StopBit,Temp);
  Erg := Erg + ',' + Temp + '  ';
  if ClearToSend then Erg := Erg + ' CTS';
  if DataSetReady then Erg := Erg + ' DSR';
  if DataCarrierDetect then Erg := Erg + ' CD';
  IdentifyComStatus := Erg;
end;


{kein Init der Unit}
end.

